@@ -37,14 +37,14 @@ impl HttpSender {
     ///
     /// The URL is an HTTP URL, usually for port 8899, as in
     /// "http://localhost:8899". The sender has a default timeout of 30 seconds.
-    pub fn new(url: String) -> Self {
+    pub fn new<U: ToString>(url: U) -> Self {
         Self::new_with_timeout(url, Duration::from_secs(30))
     }
 
     /// Create an HTTP RPC sender.
     ///
     /// The URL is an HTTP URL, usually for port 8899.
-    pub fn new_with_timeout(url: String, timeout: Duration) -> Self {
+    pub fn new_with_timeout<U: ToString>(url: U, timeout: Duration) -> Self {
         // `reqwest::blocking::Client` panics if run in a tokio async context.  Shuttle the
         // request to a different tokio thread to avoid this
         let client = Arc::new(
@@ -58,7 +58,7 @@ impl HttpSender {
 
         Self {
             client,
-            url,
+            url: url.to_string(),
             request_id: AtomicU64::new(0),
             stats: RwLock::new(RpcTransportStats::default()),
         }
